pub mod connection;
pub mod player;
pub mod server;

use std::{sync::Arc, time::Duration};

use connection::Client;
use potato_protocol::packet::clientbound;
use server::Server;
use tokio::{net::TcpListener, time::interval};

#[tokio::main]
async fn main() {
    let server = PotatoServer::new();

    server.listen_for_connections().await;
}

pub struct PotatoServer {
    server: Server,
}

impl PotatoServer {
    pub fn new() -> Self {
        PotatoServer {
            server: Server::new(),
        }
    }

    pub async fn listen_for_connections(&self) {
        let listener = TcpListener::bind("0.0.0.0:25565")
            .await
            .expect("Failed to start listening on port 25565");
        println!("listening started, ready to accept");

        loop {
            if let Ok((stream, _)) = listener.accept().await {
                let (stream_read, mut stream_send) = stream.into_split();
                let (tx, mut rx) = tokio::sync::mpsc::channel(64);
                let Ok(client) = Client::new(tx, stream_read) else {
                    println!("Failed to create client");
                    continue;
                };
                let client = Arc::new(client);

                tokio::spawn(async move {
                    while let Some(packet) = rx.recv().await {
                        if let Err(e) = packet.write_to_stream(&mut stream_send).await {
                            println!("Error while sending packet: {}", e);
                            // TODO: Close connection with client here
                            break;
                        }
                    }
                });

                // Keep alive task
                {
                    let client = client.clone();
                    tokio::spawn(async move {
                        let mut interval = interval(Duration::from_secs(10));
                        // TODO: Terminate this when the client connection is closed.
                        loop {
                            interval.tick().await;
                            if client.needs_keep_alive().await {
                                // TODO: Deal with this error by closing the connection.
                                // TODO: Use a unique id for each keep alive packet (or at least a
                                // timestamp)
                                match client
                                    .send_packet(&clientbound::KeepAlivePacket { id: 0 })
                                    .await
                                {
                                    Ok(_) => {}
                                    Err(e) => {
                                        println!("Error while sending keep alive packet: {}", e)
                                    }
                                };
                            }
                        }
                    });
                }

                {
                    let registries = self.server.registries.clone();
                    tokio::spawn(async move {
                        // TODO: Deal with errors
                        loop {
                            client
                                .read_packets()
                                .await
                                .expect("Error while reading packets");
                            client
                                .handle_packets(&registries)
                                .await
                                .expect("Error while handling packets");
                        }
                    });
                }
            }
        }
    }
}
